/* kompilator fontu pro AND-1V rev. 2.2 */
/* automaticka verze 10x12 nebo 10x24 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define X_SIZE 320               /* pocet pixelu mikroradku */
#define Y_SIZE1 96               /* pocet mikroradku pro znaky 10x12*/
#define Y_SIZE2 192              /* pocet mikroradku pro znaky 10x24*/
#define MAXITEM X_SIZE/2*Y_SIZE2 /* velikost pole obrazovych dat */
#define MAXVRAM X_SIZE/5*Y_SIZE2 /* velikost fiktivni VIDEORAM pro projekci bitmapy*/
#define BMP_HEADER_LEN 118       /* delka hlavicky souboru BMP */
#define ERR_SIZE -1              /* nepripustna velikost obrazku */
#define ERR_NUMCOL -2            /* nepripustny pocet barev obrazku */
#define ERR_FORMAT -3            /* spatny format souboru (not BMP) */

FILE *vstupni_soubor, *vystupni_soubor;
char jmeno_vstupniho_souboru[255];
char jmeno_vystupniho_souboru[255];
unsigned char hlavicka_bmp[BMP_HEADER_LEN];
unsigned char bitmapa[MAXITEM];
unsigned char videoram[MAXVRAM];
int size_x, size_y;

void napoveda(void)
{
    printf("Kompilator fontu pro kartu SAPI/AND-1V rev. 2.2\n\n");
    printf("Vstup: BMP 16 barev, rozmer 320x96 pixelu nebo 320x192\n");
    printf("Znak: 10x12 nebo 10x24 pixelu\n");
    printf("Barva aktivniho pixelu je bila (WHT), ostatni barvy ignorovany.\n");
    printf("Znaky serazeny v poctu 32 na radek, celkem 8 radku.\n");
    printf("Zkompiluje se sada 256 znaku.\n\n");
    printf("Volani: and1v22font.exe font.bmp\n\n");
    printf("STISKNI ENTER\n\n");
    getchar();
}

/* podle cisla indexu barvy nacte RGB slozky a rozpozna bilou barvu */
/* pro alespon 94% bilou vraci 1, jinak 0 */

unsigned char je_bily_pixel(unsigned char index_barvoveho_registru)
{
  unsigned char pal_adr;

  pal_adr=14+40+(index_barvoveho_registru*4); /* adresa barvoveho registru v hlavicce BMP */
  if ((hlavicka_bmp[pal_adr+0]>240) && (hlavicka_bmp[pal_adr+1]>240) && (hlavicka_bmp[pal_adr+2]>240))
    return 1;
  else
    return 0;
}

/* nacteni hodnoty pixelu z nacteneho BMP */
/* vraci hodnotu 0=cerna nebo 1=bila */

unsigned char get_pixel(int x, int y)
{
  int s_mikroradku;

  s_mikroradku = 4*(1+((size_x-1)/8)); /* interni sirka mikroradku v BMP */
  if (x < size_x)
    if (x == (x/2*2))
      return je_bily_pixel(((bitmapa[(size_y-1-y)*s_mikroradku+x/2])>>4)&0xf); /* sude pozice */
    else
      return je_bily_pixel((bitmapa[(size_y-1-y)*s_mikroradku+x/2])&0xf);      /* liche poz. */
  else
    return 0; /* mimo platna data vracim zhasnuty pixel */
}

/* nacteni hlavicky BMP a test formatu */

int analyza_dat(int *sizex, int *sizey)
{
  int ukazatel;

  ukazatel = 0;
  while (!feof(vstupni_soubor) & (ukazatel < BMP_HEADER_LEN) ) /* nacteni hlavicky BMP */
    hlavicka_bmp[ukazatel++] = (unsigned char) fgetc(vstupni_soubor);

  ukazatel = 0;
  while (!feof(vstupni_soubor) & (ukazatel < MAXITEM) ) /* nacteni bitmapy BMP */
    bitmapa[ukazatel++] = (unsigned char) fgetc(vstupni_soubor);

  /* nacteni rozmeru obrazku z hlavicky BMP souboru */
  *sizex = (unsigned long)hlavicka_bmp[18]+(unsigned long)256*(unsigned long)hlavicka_bmp[19];
  *sizey = (unsigned long)hlavicka_bmp[22]+(unsigned long)256*(unsigned long)hlavicka_bmp[23];

  if (hlavicka_bmp[0] != 'B') return ERR_FORMAT; /* not metadata BMP */
  if (hlavicka_bmp[1] != 'M') return ERR_FORMAT; /* not metadata BMP */
  if (hlavicka_bmp[28] != 4) return ERR_NUMCOL; /* not 16 barevny BMP */
  if ( (*sizex != X_SIZE) || ((*sizey != Y_SIZE1)&&(*sizey != Y_SIZE2)) ) return ERR_SIZE;

  return 0;
}

/* test formatu a nacteni bitmapy */

int bmp2data(void)
{
  int chybovy_kod,y,b,size_b,p;

  chybovy_kod = analyza_dat(&size_x, &size_y);
  if (chybovy_kod) return chybovy_kod;

  size_b = 1 + ((size_x-1)/5); /* pocet bajtu na sirku ve fiktivni videoram */

  /* vynulovat obraz videoram */
  for (y = 0; y < size_y; y++)
    for (b = 0; b < size_b; b++)
      videoram[(X_SIZE/5)*y+b] = 0;

  printf("Detekovan validni obrazek o rozmerech %i x %i pixelu.\n",size_x,size_y);

  /* vlastni prevod pixelu do obrazu videoram */
  for (y = 0; y < size_y; y++)   /* mikroradky */
    for (b = 0; b < size_b; b++) /* bajty na radku */
      for (p = 0; p < 5; p++)    /* jednotlive bity v bajtu */
                                 /* nahodit pixel, kdyz je priblizne bily */
        if (get_pixel(5*b+p,y)==1) videoram[(X_SIZE/5)*y+b] |= (1<<(4-p));
  return 0;
}

/* otoceni poradi bitu a jejich ustaveni do pozice */

unsigned char mirror_byte(unsigned char native)
{
  unsigned char pom;
  pom = 0;
  pom |= (native >> 1) & 8;
  pom |= (native << 1) & 16;
  pom |= (native << 3) & 32;
  pom |= (native << 5) & 64;
  pom |= (native << 7) & 128;
  return pom;
}

/* data z obrazu bitmapy 320x96 preklopi do BIN podoby */
/* EEPROM vyuziva jen bity D7:D3 */

void data2bin_1012(void)
{
  unsigned int polovina, index, znak_x, znak_y;

  for (polovina=0; polovina<2; polovina++)    /* L a P polovina fontu po 5 pixelech */
    for (znak_y = 0; znak_y < 8; znak_y++)    /* 8 radku fontu */
      for (znak_x = 0; znak_x < 32; znak_x++) /* 32 znaku na radku */
        for (index = 0; index < 16; index++)  /* 16 adresovych pozic pro jeden znak */
          switch(index)
            {
              case 0:
              case 1:
              case 8:
              case 9:  fputc(0,vystupni_soubor);
                       break;
              case 2:
              case 3:
              case 4:
              case 5:
              case 6:
              case 7:  fputc(mirror_byte(videoram[((znak_y*12)+(index-2))*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                       break;
              case 10:
              case 11:
              case 12:
              case 13:
              case 14:
              case 15: fputc(mirror_byte(videoram[((znak_y*12)+(index-4))*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                       break;
            }
}

/* data z obrazu bitmapy 320x192 preklopi do BIN podoby */
/* EEPROM vyuziva jen bity D7:D3 */

void data2bin_1024(void)
{
  unsigned int polovina, pageK32, index, znak_x, znak_y;

  for (pageK32=0; pageK32<4; pageK32++)         /* 8k stranky ve vetsich pametech 27C128, 27C256 a 28C256 */
    for (polovina=0; polovina<2; polovina++)    /* L a P polovina fontu po 5 pixelech */
      for (znak_y = 0; znak_y < 8; znak_y++)    /* 8 radku fontu */
        for (znak_x = 0; znak_x < 32; znak_x++) /* 32 znaku na radku */
          for (index = 0; index < 16; index++)  /* 16 adresovych pozic pro jeden znak */
            switch(index)
              {
                case 0:
                case 1:
                case 8:
                case 9:  fputc(0,vystupni_soubor);
                         break;
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                case 7:  if ((pageK32==0)||(pageK32==2))
                           fputc(mirror_byte(videoram[((znak_y*24)+(2*(index-2)))*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                         else
                           fputc(mirror_byte(videoram[((znak_y*24)+(2*(index-2))+1)*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                         break;
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15: if ((pageK32==0)||(pageK32==2))
                           fputc(mirror_byte(videoram[((znak_y*24)+(2*(index-4)))*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                         else
                           fputc(mirror_byte(videoram[((znak_y*24)+(2*(index-4))+1)*(X_SIZE/5)+(2*znak_x+polovina)]),vystupni_soubor);
                         break;
              }
}

/* hlavni procedura konvertoru */

int main(int argc, char *argv[])
{
    if (argc<2)
      napoveda();
    else
    {
      strcpy(jmeno_vstupniho_souboru,argv[1]);
      strncpy(jmeno_vystupniho_souboru,jmeno_vstupniho_souboru,strlen(jmeno_vstupniho_souboru)-3);
      strcat(jmeno_vystupniho_souboru,"bin");
      vstupni_soubor = fopen(jmeno_vstupniho_souboru, "rb");
      if (vstupni_soubor == NULL)
        printf("Soubor %s nenalezen!",jmeno_vstupniho_souboru);
      else
        {
          vystupni_soubor = fopen(jmeno_vystupniho_souboru, "wb");
          if (vystupni_soubor == NULL)
            printf("Vystupni soubor nelze otevrit!");
          else
            {
              switch (bmp2data())
                {
                  case ERR_SIZE:   printf("Obrazek nema povolenou velikost 320x96 (320x192) pixelu.\n");
                                   printf("Prevod nebude proveden.");
                                   break;
                  case ERR_NUMCOL: printf("Vstupni soubor nema format 16 barevneho BMP.\n");
                                   printf("Prevod nebude proveden.");
                                   break;
                  case ERR_FORMAT: printf("Nepripustny format vstupniho souboru.\n");
                                   printf("Prevod nebude proveden.");
                                   break;
                  case 0:          if (size_y==Y_SIZE1)
                                     data2bin_1012();  /* prevod rastru 10x12 na BIN */
                                   else
                                     data2bin_1024();  /* prevod rastru 10x24 na BIN */
                                   break;
                  default: printf("Nespecifikovana chyba pri prevodu dat.\n");
                           printf("Prevod nebude proveden.");
                           break;
                }
              fclose(vystupni_soubor);
            }
          fclose(vstupni_soubor);
        }
    printf("\n\nStiskni ENTER\n\n\n");
    getchar();
  }
  return 0;
}
